from textual.content import Content
from textual.widgets import Static


async def test_content_property():
    static = Static()
    assert static.content == ""
    static.content = "Foo"
    assert static.content == "Foo"
    assert isinstance(static.content, str)
    assert static.visual == "Foo"
    assert isinstance(static.visual, Content)

    static.update("Hello")
    assert static.content == "Hello"
    assert isinstance(static.content, str)
    assert static.visual == "Hello"
    assert isinstance(static.visual, Content)



import pytest

from textual.app import App, ComposeResult
from textual.geometry import Offset
from textual.selection import Selection
from textual.widgets import Static


@pytest.mark.parametrize(
    "text,selection,expected",
    [
        ("Hello", Selection(None, None), "Hello"),
        ("Hello\nWorld", Selection(None, None), "Hello\nWorld"),
        ("Hello\nWorld", Selection(Offset(0, 1), None), "World"),
        ("Hello\nWorld", Selection(None, Offset(5, 0)), "Hello"),
        ("Foo", Selection(Offset(0, 0), Offset(1, 0)), "F"),
        ("Foo", Selection(Offset(1, 0), Offset(2, 0)), "o"),
        ("Foo", Selection(Offset(0, 0), Offset(2, 0)), "Fo"),
        ("Foo", Selection(Offset(0, 0), None), "Foo"),
    ],
)
def test_extract(text: str, selection: Selection, expected: str) -> None:
    """Test Selection.extract"""
    assert selection.extract(text) == expected


async def test_double_width():
    """Test that selection works with double width characters."""

    TEXT = """😂❤️👍Select😊🙏😍\nme🔥💯😭😂❤️👍"""

    class TextSelectApp(App):
        def compose(self) -> ComposeResult:
            yield Static(TEXT)

    app = TextSelectApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        assert await pilot.mouse_down(offset=(2, 0))
        await pilot.pause()
        assert await pilot.mouse_up(offset=(7, 1))
        selected_text = app.screen.get_selected_text()
        expected = "❤️👍Select😊🙏😍\nme🔥💯😭"

    assert selected_text == expected



from rich.color import Color as RichColor
from rich.color import ColorType
from rich.segment import Segment
from rich.style import Style
from rich.terminal_theme import MONOKAI

from textual.color import Color
from textual.filter import ANSIToTruecolor


def test_ansi_to_truecolor_8_bit_dim():
    """Test that converting an 8-bit color with dim doesn't crash.

    Regression test for https://github.com/Textualize/textual/issues/5946

    """
    # Given
    ansi_filter = ANSIToTruecolor(MONOKAI)
    test_color = RichColor("color(253)", ColorType.EIGHT_BIT, number=253)
    test_style = Style(color=test_color, dim=True)
    segments = [Segment("This should not crash", style=test_style)]
    background_color = Color(0, 0, 0)

    # When
    # This line will crash if the bug is present
    new_segments = ansi_filter.apply(segments, background_color)

    # Then
    assert new_segments is not None



from textual.app import App, ComposeResult
from textual.containers import Middle
from textual.widgets import Placeholder


class Box(Placeholder):
    """Example widget."""

    DEFAULT_CSS = """
    Box {
        width: 16;
        height: 5;        
    }
    """


class ContainerApp(App):
    """Simple app to play with containers."""

    CSS = """
    .with-border {
        border: heavy green;
    }
    """

    def compose(self) -> ComposeResult:
        with Middle(classes="with-border"):  # (1)!
            yield Box("Box 1.")
            yield Box("Box 2.")
            yield Box("Box 3.")


if __name__ == "__main__":
    app = ContainerApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Markdown

EXAMPLE_MARKDOWN = """\
## Markdown

- Typography *emphasis*, **strong**, `inline code` etc.    
- Headers    
- Lists    
- Syntax highlighted code blocks
- Tables and more

## Quotes

> I must not fear.
> > Fear is the mind-killer.
> > Fear is the little-death that brings total obliteration.
> > I will face my fear.
> > > I will permit it to pass over me and through me.
> > > And when it has gone past, I will turn the inner eye to see its path.
> > > Where the fear has gone there will be nothing. Only I will remain.

## Tables

| Name            | Type   | Default | Description                        |
| --------------- | ------ | ------- | ---------------------------------- |
| `show_header`   | `bool` | `True`  | Show the table header              |
| `fixed_rows`    | `int`  | `0`     | Number of fixed rows               |
| `fixed_columns` | `int`  | `0`     | Number of fixed columns            |

## Code blocks

```python
def loop_last(values: Iterable[T]) -> Iterable[Tuple[bool, T]]:
    \"\"\"Iterate and generate a tuple with a flag for last value.\"\"\"
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    for value in iter_values:
        yield False, previous_value
        previous_value = value
    yield True, previous_value
```
    

"""


class MarkdownExampleApp(App):

    def compose(self) -> ComposeResult:
        markdown = Markdown(EXAMPLE_MARKDOWN)
        markdown.code_indent_guides = False
        yield markdown


if __name__ == "__main__":
    app = MarkdownExampleApp()
    app.run()



import inspect
from typing import Any

from textual import work
from textual._log import LogGroup, LogVerbosity
from textual.app import App


async def test_log_from_worker() -> None:
    """Check that log calls from threaded workers call app._log"""

    log_messages: list[tuple] = []

    class LogApp(App):

        def _log(
            self,
            group: LogGroup,
            verbosity: LogVerbosity,
            _textual_calling_frame: inspect.Traceback,
            *objects: Any,
            **kwargs,
        ) -> None:
            log_messages.append(objects)

        @property
        def _is_devtools_connected(self):
            """Fake connected devtools."""
            return True

        def on_mount(self) -> None:
            self.do_work()
            self.call_after_refresh(self.exit)

        @work(thread=True)
        def do_work(self) -> None:
            self.log("HELLO from do_work")

    app = LogApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        assert ("HELLO from do_work",) in log_messages



import pytest

from textual.content import Span
from textual.highlight import guess_language, highlight


def test_highlight() -> None:
    """Test simple application of highlight."""
    import_this = highlight("import this", language="python")
    assert import_this.plain == "import this"
    print(import_this.spans)
    assert import_this.spans == [
        Span(0, 11, style="$text"),
        Span(0, 6, style="$text-error"),
        Span(7, 11, style="$text-primary"),
    ]


@pytest.mark.parametrize(
    "code,path,language",
    [
        ("", "", "default"),
        ("# Don't matter", "foo.tcss", "scss"),
        ("import this", "foo.py", "python"),
        ("<xml>", "foo.xml", "xml"),
        ("{}", "data.json", "json"),
        ("#! python", "", "python"),
        ("", "foo.py", "python"),
    ],
)
def test_guess_language(code: str, path: str, language: str) -> None:
    """Test guess_language is working."""
    assert guess_language(code, path) == language



from __future__ import annotations

import pytest
from rich.text import Text

from textual.color import Color
from textual.content import Content, Span
from textual.style import Style
from textual.visual import RenderOptions
from textual.widget import Widget


def test_blank():
    """Check blank content."""
    blank = Content("")
    assert isinstance(blank, Content)
    assert str(blank) == ""
    assert blank.plain == ""
    assert not blank
    assert blank.markup == ""
    assert len(blank) == 0
    assert blank.spans == []


def test_simple():
    """Check content with simple unstyled text."""
    simple = Content("foo")
    assert isinstance(simple, Content)
    assert str(simple) == "foo"
    assert simple.plain == "foo"
    assert simple  # Not empty is truthy
    assert simple.markup == "foo"
    assert len(simple) == 3
    assert simple.spans == []


def test_constructor():
    content = Content("Hello, World")
    assert content
    assert len(content) == 12
    assert content.cell_length == 12
    assert content.plain == "Hello, World"
    repr(content)


def test_bool():
    assert bool(Content("foo")) is True
    assert bool(Content("")) is False


def test_from_rich_text():
    text = Text.from_markup("[red]Hello[/red] [blue]World[/blue]")
    content = Content.from_rich_text(text)
    assert len(content) == 11
    assert content.plain == "Hello World"
    print(content.spans)
    assert content.spans == [
        Span(0, 5, style=Style(foreground=Color(128, 0, 0, ansi=1))),
        Span(6, 11, style=Style(foreground=Color(0, 0, 128, ansi=4))),
    ]


def test_styled():
    text = Content.styled("Hello", "red")
    assert text.plain == "Hello"
    assert len(text) == 5
    assert text.cell_length == 5
    assert text._spans == [Span(0, 5, "red")]


def test_getitem():
    content = Content("Hello, world").stylize("blue", 0, 5)
    assert content[0].plain == "H"
    assert content[0]._spans == [Span(0, 1, "blue")]
    assert content[-1].plain == "d"
    assert content[-1]._spans == []
    assert content[:2].plain == "He"
    assert content[:2]._spans == [Span(0, 2, "blue")]


def test_cell_length():
    assert Content("").cell_length == 0
    assert Content("foo").cell_length == 3
    assert Content("💩").cell_length == 2


def test_stylize() -> None:
    """Test the stylize method."""
    foo = Content("foo bar")
    assert foo.spans == []
    red_foo = foo.stylize("red")
    # stylize create a new object
    assert foo.spans == []
    # With no parameters, full string is stylized
    assert red_foo.spans == [Span(0, 7, "red")]
    red_foo = red_foo.stylize("blue", 4, 7)
    # Another span is appended
    assert red_foo.spans == [
        Span(0, 7, "red"),
        Span(4, 7, "blue"),
    ]


def test_stylize_before() -> None:
    """Test the stylize_before method."""
    foo = Content("foo bar")
    assert foo.spans == []
    red_foo = foo.stylize("red")
    # stylize create a new object
    assert foo.spans == []
    # With no parameters, full string is stylized
    assert red_foo.spans == [Span(0, 7, "red")]
    red_foo = red_foo.stylize_before("blue", 4, 7)
    # Another span is appended
    assert red_foo.spans == [
        Span(4, 7, "blue"),
        Span(0, 7, "red"),
    ]


def test_eq() -> None:
    """Test equality."""
    assert Content("foo") == Content("foo")
    assert Content("foo") == "foo"
    assert Content("foo") != Content("bar")
    assert Content("foo") != "bar"


def test_add() -> None:
    """Test addition."""
    # Simple cases
    assert Content("") + Content("") == Content("")
    assert Content("foo") + Content("") == Content("foo")
    # Works with simple strings
    assert Content("foo") + "" == Content("foo")
    assert "" + Content("foo") == Content("foo")

    # Test spans after addition
    content = Content.styled("foo", "red") + " " + Content.styled("bar", "blue")
    assert str(content) == "foo bar"
    assert content.spans == [Span(0, 3, "red"), Span(4, 7, "blue")]
    assert content.cell_length == 7


def test_radd() -> None:
    """Test reverse addition."""
    assert "foo" + Content("bar") == Content("foobar")

    # Test spans after addition
    content = "foo " + Content.styled("bar", "blue")
    assert str(content) == "foo bar"
    assert content.spans == [Span(4, 7, "blue")]


def test_from_markup():
    """Test simple parsing of content markup."""
    content = Content.from_markup("[red]Hello[/red] [blue]World[/blue]")
    assert len(content) == 11
    assert content.plain == "Hello World"
    assert content.spans == [
        Span(start=0, end=5, style="red"),
        Span(start=6, end=11, style="blue"),
    ]


def test_markup():
    """Test markup round trip"""
    content = Content.from_markup("[red]Hello[/red] [blue]World[/blue]")
    assert content.plain == "Hello World"
    assert content.markup == "[red]Hello[/red] [blue]World[/blue]"


def test_join():
    """Test the join method."""

    # Edge cases
    assert Content("").join([]) == ""
    assert Content(".").join([]) == ""
    assert Content("").join(["foo"]) == "foo"
    assert Content(".").join(["foo"]) == "foo"

    # Join strings and Content
    pieces = [Content.styled("foo", "red"), "bar", Content.styled("baz", "blue")]
    content = Content(".").join(pieces)
    assert content.plain == "foo.bar.baz"
    assert content.spans == [Span(0, 3, "red"), Span(8, 11, "blue")]


def test_sort():
    """Test content may be sorted."""
    # functools.total_ordering doing most of the heavy lifting here.
    contents = sorted([Content("foo"), Content("bar"), Content("baz")])
    assert contents[0].plain == "bar"
    assert contents[1].plain == "baz"
    assert contents[2].plain == "foo"


def test_truncate():
    """Test truncated method."""
    content = Content.from_markup("[red]Hello World[/red]")
    # Edge case of 0
    assert content.truncate(0).markup == ""
    # Edge case of 0 wil ellipsis
    assert content.truncate(0, ellipsis=True).markup == ""
    # Edge case of 1
    assert content.truncate(1, ellipsis=True).markup == "[red]…[/red]"
    # Truncate smaller
    assert content.truncate(3).markup == "[red]Hel[/red]"
    # Truncate to same size
    assert content.truncate(11).markup == "[red]Hello World[/red]"
    # Truncate smaller will ellipsis
    assert content.truncate(5, ellipsis=True).markup == "[red]Hell…[/red]"
    # Truncate larger results unchanged
    assert content.truncate(15).markup == "[red]Hello World[/red]"
    # Truncate larger with padding increases size
    assert content.truncate(15, pad=True).markup == "[red]Hello World[/red]    "


def test_assemble():
    """Test Content.assemble constructor."""
    content = Content.assemble(
        "Assemble: ",  # Simple text
        Content.from_markup(
            "pieces of [red]text[/red] or [blue]content[/blue] into "
        ),  # Other Content
        ("a single Content instance", "underline"),  # A tuple of text and a style
    )
    assert (
        content.plain
        == "Assemble: pieces of text or content into a single Content instance"
    )
    print(content.spans)
    assert content.spans == [
        Span(20, 24, style="red"),
        Span(28, 35, style="blue"),
        Span(41, 66, style="underline"),
    ]


@pytest.mark.parametrize(
    "markup,plain",
    [
        ("\\[", "["),
        ("\\[foo", "[foo"),
        ("\\[foo]", "[foo]"),
        ("\\[/foo", "[/foo"),
        ("\\[/foo]", "[/foo]"),
        ("\\[]", "[]"),
        ("\\[0]", "[0]"),
    ],
)
def test_escape(markup: str, plain: str) -> None:
    """Test that escaping the first square bracket."""
    content = Content.from_markup(markup)
    assert content.plain == plain
    assert content.spans == []


def test_strip_control_codes():
    """Test that control codes are removed from content."""
    assert Content("foo\r\nbar").plain == "foo\nbar"
    assert Content.from_markup("foo\r\nbar").plain == "foo\nbar"


def test_content_from_text():
    # Check markup
    content1 = Content.from_text("[bold]Hello World")
    assert isinstance(content1, Content)
    assert content1.plain == "Hello World"
    assert content1.spans == [Span(0, 11, "bold")]

    # Check Content is unaltered
    content2 = Content.from_text(content1)
    assert content1 is content2

    # Check Text instance is converted to Content
    content3 = Content.from_text(Text.from_markup("[bold]Hello World"))
    assert isinstance(content3, Content)
    assert content3.plain == "Hello World"
    assert content3.spans == [Span(0, 11, Style(bold=True))]

    # Check markup disabled
    content4 = Content.from_text("[bold]Hello World", markup=False)
    assert isinstance(content4, Content)
    assert content4.plain == "[bold]Hello World"
    assert content4.spans == []


def test_first_line():
    content = Content("foo\nbar").stylize("red")
    first_line = content.first_line
    assert first_line.plain == "foo"
    assert first_line.spans == [Span(0, 3, "red")]


async def test_split_and_tabs():
    spans = [
        Span(0, 49, style="$text"),
    ]

    content = Content("--- hello.py\t2024-01-15 10:30:00.000000000 -0800", spans=spans)
    widget = Widget()
    content.render_strips(0, None, Style(), RenderOptions(widget._get_style, {}))


def test_simplify():
    """Test simplify joins spans."""
    content = Content.from_markup("[bold]Foo[/][bold]Bar[/]")
    assert content.spans == [Span(0, 3, "bold"), Span(3, 6, "bold")]
    content.simplify()
    assert content.spans == [Span(0, 6, "bold")]


@pytest.mark.parametrize(
    ["input", "tab_width", "expected"],
    [
        (Content(""), 8, Content("")),
        (Content("H"), 8, Content("H")),
        (Content("Hello"), 8, Content("Hello")),
        (Content("\t"), 8, Content(" " * 8)),
        (Content("A\t"), 8, Content("A" + " " * 7)),
        (Content("ABCD\t"), 8, Content("ABCD" + " " * 4)),
        (Content("ABCDEFG\t"), 8, Content("ABCDEFG ")),
        (Content("ABCDEFGH\t"), 8, Content("ABCDEFGH" + " " * 8)),
        (Content("Hel\tlo!"), 4, Content("Hel lo!")),
        (Content("\t\t"), 4, Content(" " * 8)),
        (Content("FO\t\t"), 4, Content("FO      ")),
        (Content("FO\tOB\t"), 4, Content("FO  OB  ")),
        (
            Content("FOO", spans=[Span(0, 3, "red")]),
            4,
            Content("FOO", spans=[Span(0, 3, "red")]),
        ),
        (
            Content("FOO\tBAR", spans=[Span(0, 3, "red")]),
            8,
            Content("FOO     BAR", spans=[Span(0, 3, "red")]),
        ),
        (
            Content("FOO\tBAR", spans=[Span(0, 3, "red"), Span(4, 8, "blue")]),
            8,
            Content("FOO     BAR", spans=[Span(0, 3, "red"), Span(8, 11, "blue")]),
        ),
        (
            Content("foo\tbar\nbaz", spans=[Span(0, 11, "red")]),
            8,
            Content("foo     bar\nbaz", spans=[Span(0, 15, "red")]),
        ),
    ],
)
def test_expand_tabs(input: Content, tab_width: int, expected: Content):
    output = input.expand_tabs(tab_width).simplify()
    print(repr(output))
    assert output.plain == expected.plain
    assert output._spans == expected._spans


def test_add_spans() -> None:
    content = Content.from_markup("[red]Hello[/red], World!")
    content = content.add_spans([Span(0, 5, "green"), Span(7, 9, "blue")])
    expected = [
        Span(0, 5, style="red"),
        Span(0, 5, style="green"),
        Span(7, 9, style="blue"),
    ]
    assert content.spans == expected


def test_wrap() -> None:
    content = Content.from_markup("[green]Hello, [b]World, One two three[/b]")
    wrapped = content.wrap(6)
    print(wrapped)
    expected = [
        Content("Hello,", spans=[Span(0, 6, style="green")]),
        Content("World,", spans=[Span(0, 6, style="green"), Span(0, 6, style="b")]),
        Content("One", spans=[Span(0, 3, style="green"), Span(0, 3, style="b")]),
        Content("two", spans=[Span(0, 3, style="green"), Span(0, 3, style="b")]),
        Content("three", spans=[Span(0, 5, style="green"), Span(0, 5, style="b")]),
    ]
    assert len(wrapped) == len(expected)
    for line1, line2 in zip(wrapped, expected):
        assert line1.is_same(line2)


@pytest.mark.parametrize(
    "content, width, expected",
    [
        (
            Content("111222333"),
            3,
            [
                Content("111"),
                Content("222"),
                Content("333"),
            ],
        ),
        (
            Content("1112223334"),
            3,
            [
                Content("111"),
                Content("222"),
                Content("333"),
                Content("4"),
            ],
        ),
        (
            Content(""),
            10,
            [Content("")],
        ),
        (
            Content("1"),
            10,
            [Content("1")],
        ),
        (
            Content("📦"),
            10,
            [Content("📦")],
        ),
        (
            Content("📦"),
            1,
            [Content("📦")],
        ),
        (
            Content("Hello"),
            10,
            [Content("Hello")],
        ),
        (
            Content("Hello"),
            5,
            [Content("Hello")],
        ),
        (
            Content("Hello"),
            2,
            [Content("He"), Content("ll"), Content("o")],
        ),
        (
            Content.from_markup("H[b]ell[/]o"),
            2,
            [
                Content.from_markup("H[b]e"),
                Content.from_markup("[b]ll[/]"),
                Content("o"),
            ],
        ),
        (
            Content.from_markup("💩H[b]ell[/]o"),
            2,
            [
                Content("💩"),
                Content.from_markup("H[b]e"),
                Content.from_markup("[b]ll[/]"),
                Content("o"),
            ],
        ),
        (
            Content.from_markup("💩H[b]ell[/]o"),
            3,
            [
                Content("💩H"),
                Content.from_markup("[b]ell"),
                Content.from_markup("o"),
            ],
        ),
        (
            Content.from_markup("💩H[b]ell[/]o💩"),
            3,
            [
                Content("💩H"),
                Content.from_markup("[b]ell"),
                Content.from_markup("o💩"),
            ],
        ),
        (
            Content.from_markup("💩💩💩"),
            1,
            [
                Content("💩"),
                Content("💩"),
                Content("💩"),
            ],
        ),
        (
            Content.from_markup("💩💩💩"),
            3,
            [
                Content("💩"),
                Content("💩"),
                Content("💩"),
            ],
        ),
        (
            Content.from_markup("💩💩💩"),
            4,
            [
                Content("💩💩"),
                Content("💩"),
            ],
        ),
        (
            Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999"),
            50,
            [Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999")],
        ),
        (
            Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999"),
            49,
            [
                Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦99"),
                Content("9"),
            ],
        ),
        (
            Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999"),
            48,
            [
                Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦9"),
                Content("99"),
            ],
        ),
        (
            Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999"),
            47,
            [
                Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦"),
                Content("999"),
            ],
        ),
        (
            Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999"),
            46,
            [
                Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888"),
                Content("📦999"),
            ],
        ),
        (
            Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999"),
            45,
            [
                Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888"),
                Content("📦999"),
            ],
        ),
        (
            Content("📦000📦111📦222📦333📦444📦555📦666📦777📦888📦999"),
            44,
            [
                Content("📦000📦111📦222📦333📦444📦555📦666📦777📦88"),
                Content("8📦999"),
            ],
        ),
    ],
)
def test_fold(content: Content, width: int, expected: list[Content]) -> None:
    """Test content.fold method works, and correctly handles double width cells.

    Args:
        content: Test content.
        width: Desired width.
        expected: Expectected result.
    """
    result = content.fold(width)
    assert isinstance(result, list)
    assert len(result) == len(expected)
    for line, expected_line in zip(result, expected):
        assert line.is_same(expected_line)


@pytest.mark.parametrize(
    "width,style,text,spans,cell_length",
    [
        (5, None, "     ", [], 5),
        (0, None, "", [], 0),
        (5, "on red", "     ", [Span(0, 5, "on red")], 5),
    ],
)
def test_blank_method(
    width: int, style: str | None, text: str, spans: list[Span], cell_length: int
) -> None:
    blank = Content.blank(width, style)
    assert blank.plain == text
    assert blank.spans == spans
    assert blank.cell_length == cell_length



from textual import containers, widgets
from textual.app import App, ComposeResult
from textual.layouts.grid import GridLayout


async def test_grid_size():
    """Test the `grid_size` property on GridLayout."""

    class GridApp(App):
        CSS = """
        Grid {
            grid-size: 3;
            grid-columns: auto;
            height: auto;
            Label {
                padding: 2 4;
                border: blue;
            }
        }
        """

        def compose(self) -> ComposeResult:
            with containers.VerticalScroll():
                with containers.Grid():
                    for _ in range(7):
                        yield widgets.Label("Hello, World!")

    app = GridApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        await app.wait_for_refresh()
        grid_layout = app.query_one(containers.Grid).layout
        assert isinstance(grid_layout, GridLayout)
        assert grid_layout.grid_size == (3, 3)



import pytest

from textual import containers, getters
from textual.app import App, ComposeResult
from textual.css.query import NoMatches, WrongType
from textual.widget import Widget
from textual.widgets import Input, Label


async def test_getters() -> None:
    """Check the getter descriptors work, and return expected errors."""

    class QueryApp(App):
        label1 = getters.query_one("#label1", Label)
        label2 = getters.child_by_id("label2", Label)
        label1_broken = getters.query_one("#label1", Input)
        label2_broken = getters.child_by_id("label2", Input)
        label1_missing = getters.query_one("#foo", Widget)
        label2_missing = getters.child_by_id("bar", Widget)

        def compose(self) -> ComposeResult:
            with containers.Vertical():
                yield Label(id="label1", classes="red")
            yield Label(id="label2", classes="green")

    app = QueryApp()
    async with app.run_test():

        assert isinstance(app.label1, Label)
        assert app.label1.id == "label1"

        assert isinstance(app.label2, Label)
        assert app.label2.id == "label2"

        with pytest.raises(WrongType):
            app.label1_broken

        with pytest.raises(WrongType):
            app.label2_broken

        with pytest.raises(NoMatches):
            app.label1_missing

        with pytest.raises(NoMatches):
            app.label2_missing


async def test_app_getter() -> None:

    class MyApp(App):
        def compose(self) -> ComposeResult:
            my_widget = MyWidget()
            my_widget.app
            yield my_widget

    class MyWidget(Widget):
        app = getters.app(MyApp)

    app = MyApp()
    async with app.run_test():
        assert isinstance(app.query_one(MyWidget).app, MyApp)
